# GERBLS version
__version__ = "0.6.4"

# Compiled Cython library
from _gerbls import *

# Core GERBLS functionality
from .blsfunc import run_bls

# Optional extras
try:
    from .clean import clean_savgol
except ImportError:
    pass


# cython: language_level = 3
# BLS model and analyzer to be included in gerbls.pyx

cdef class pyBLSModel:
    cdef BLSModel* cPtr
    cdef bool_t alloc           # Whether responsible for memory allocation
    
    def __cinit__(self):
        if type(self) is pyBLSModel:
            self.alloc = False
    
    def __dealloc__(self):
        if self.alloc and type(self) is pyBLSModel:
            del self.cPtr
    
    @property
    def freq(self):
        return np.asarray(self.view_freq())
    
    #def get_max_duration(self, double P):
    #    return self.cPtr.get_max_duration(P)
    
    @property
    def N_freq(self):
        return self.cPtr.N_freq()
    
    def run(self, bool_t verbose = False):
        self.cPtr.run(verbose)

    cdef size_t [::1] view_bins(self):
        return <size_t [:self.N_freq]>self.cPtr.N_bins.data()
        
    cdef double [::1] view_dchi2(self):
        return <double [:self.N_freq]>self.cPtr.dchi2.data()
    
    cdef double [::1] view_dmag(self):
        return <double [:self.N_freq]>self.cPtr.chi2_dmag.data()
    
    cdef double [::1] view_dur(self):
        return <double [:self.N_freq]>self.cPtr.chi2_dt.data()
    
    cdef double [::1] view_freq(self):
        return <double [:self.N_freq]>self.cPtr.freq.data()
    
    cdef double [::1] view_mag0(self):
        return <double [:self.N_freq]>self.cPtr.chi2_mag0.data()
    
    cdef double [::1] view_t0(self):
        return <double [:self.N_freq]>self.cPtr.chi2_t0.data()

cdef class pyBruteForceBLS(pyBLSModel):
    cdef BLSModel_bf* dPtr
    
    def __cinit__(self):
        self.alloc = False
    
    def __dealloc__(self):
        if self.alloc:
            del self.dPtr
    
    def setup(self,
              pyDataContainer data not None,
              double min_period,
              double max_period,
              pyTarget target = None,
              double dt_per_step = 0.,
              double t_bins = 0.,
              size_t N_bins_min = 0,
              str duration_mode = "",
              double min_duration_factor = 0.,
              double max_duration_factor = 0.):
        cdef Target* targetPtr = (<Target *>NULL if target == None else target.cPtr)
        self.dPtr = new BLSModel_bf(data.cPtr[0],
                                    1/max_period,
                                    1/min_period,
                                    targetPtr, 
                                    dt_per_step,
                                    t_bins,
                                    N_bins_min,
                                    convert_duration_mode(duration_mode),
                                    min_duration_factor,
                                    max_duration_factor)
        self.cPtr = self.dPtr
        self.alloc = True
    
    # Setup with a pre-defined frequency array
    def setup_from_freq(self,
                        pyDataContainer data not None,
                        double[:] freq_,
                        pyTarget target = None,
                        double t_bins = 0.,
                        size_t N_bins_min = 0,
                        str duration_mode = "",
                        double min_duration_factor = 0.,
                        double max_duration_factor = 0.):
        cdef Target* targetPtr = (<Target *>NULL if target == None else target.cPtr)
        self.dPtr = new BLSModel_bf(data.cPtr[0],
                                    list(freq_),
                                    targetPtr,
                                    t_bins,
                                    N_bins_min,
                                    convert_duration_mode(duration_mode),
                                    min_duration_factor,
                                    max_duration_factor)
        self.cPtr = self.dPtr
        self.alloc = True

cdef class pyFastBLS(pyBLSModel):
    cdef BLSModel_FFA* dPtr
    
    def __cinit__(self):
        self.alloc = False
    
    def __dealloc__(self):
        if self.alloc:
            del self.dPtr
    
    @property
    def dchi2(self):
        snr = np.asarray(<double [:self.dPtr.snr.size()]>self.dPtr.snr.data())
        N_widths = self.dPtr.widths.size()
        return snr.reshape((int(len(snr) / N_widths), N_widths))
    
    @property
    def foldbins(self):
        return np.asarray(<size_t [:self.dPtr.foldbins.size()]>self.dPtr.foldbins.data())
    
    @property
    def periods(self):
        return np.asarray(<double [:self.dPtr.periods.size()]>self.dPtr.periods.data())
    
    @property
    def rdata(self):
        return pyDataContainer.from_ptr(self.dPtr.rdata.get(), False)
    
    def run_double(self, bool_t verbose = True):
        self.dPtr.run_double(verbose)
    
    def setup(self,
              pyDataContainer data,
              double min_period,
              double max_period,
              pyTarget target = None,
              double t_samp = 0.,
              bool_t verbose = True,
              str duration_mode = "",
              double min_duration_factor = 0.,
              double max_duration_factor = 0.):
        cdef Target* targetPtr = (<Target *>NULL if target == None else target.cPtr)
        self.dPtr = new BLSModel_FFA(data.cPtr[0],
                                     1./max_period,
                                     1./min_period,
                                     targetPtr,
                                     convert_duration_mode(duration_mode),
                                     min_duration_factor,
                                     max_duration_factor)
        if t_samp > 0:
            self.t_samp = t_samp
        else:
            self.t_samp = np.median(np.diff(data.rjd))
            if verbose:
                print(
                    f"BLS time sampling set to the median cadence of input data: "
                    f"{self.t_samp*24*60:.2f} minutes.",
                    flush=True)
        self.cPtr = self.dPtr
        self.alloc = True
    
    @property
    def t_samp(self):
        return self.dPtr.t_samp
    @t_samp.setter
    def t_samp(self, double value):
        self.dPtr.t_samp = value
        
    @property
    def t_widths(self):
        return np.asarray(<size_t [:self.dPtr.widths.size()]>self.dPtr.widths.data())
    
    @property
    def t0(self):
        t0 = np.asarray(<size_t [:self.dPtr.t0.size()]>self.dPtr.t0.data())
        N_widths = self.dPtr.widths.size()
        return t0.reshape((int(len(t0) / N_widths), N_widths))

cdef class pyBLSAnalyzer:
    cdef size_t [:] _bins
    cdef double [:] _dchi2
    cdef double [:] _dmag
    cdef double [:] _dur
    cdef double [:] _freq
    cdef double [:] _mag0
    cdef bool_t [:] _mask
    cdef double [:] _t0
    cdef int N_freq
    cdef double t_samp
    
    def __cinit__(self, pyBLSModel model):
        self._bins = model.view_bins()
        self._dchi2 = model.view_dchi2()
        self._dmag = model.view_dmag()
        self._dur = model.view_dur()
        self._freq = model.view_freq()
        self._mag0 = model.view_mag0()
        self._t0 = model.view_t0()
        self.N_freq = model.N_freq
        self.t_samp = (model.t_samp if hasattr(model, "t_samp") else 0)
        self.initialize_mask()
    
    @property
    def dchi2(self):
        return np.asarray(self._dchi2)
    
    @property
    def dmag(self):
        return np.asarray(self._dmag)
    
    @property
    def dur(self):
        return np.asarray(self._dur)
    
    @property
    def f(self):
        return np.asarray(self._freq)
    
    cdef void initialize_mask(self):
        self._mask = np.ones(self.N_freq, dtype = np.bool_)
        # Ignore anti-transits
        self._mask *= (self.dmag > 0)
        
    @property
    def mag0(self):
        return np.asarray(self._mag0)
    
    @property
    def mask(self):
        return np.asarray(self._mask)

    @property
    def N_bins(self):
        return np.asarray(self._bins)
    
    @property
    def P(self):
        return self.f**-1
    
    @property
    def t0(self):
        return np.asarray(self._t0)
    
    def generate_models(self, N_models, double unmaskf = 0.005):
        self.initialize_mask()
        return [self.generate_next_model(unmaskf) for _ in range(N_models)]
    
    def generate_next_model(self, double unmaskf = 0.005):
        
        if not self.mask.any():
            return None
        
        cdef size_t mask_index = np.argmax(-self.dchi2[self.mask])
        cdef size_t index = np.where(self.mask)[0][mask_index]
        
        # Returned frequencies must be some range apart
        self.unmask_freq(self._freq[index], unmaskf)
        
        return pyBLSResult(self, index)
    
    # Mask out BLS frequencies less than df away from f_
    cpdef void unmask_freq(self, double f_, double df):
        self._mask *= (np.abs(self.f - f_) >= df)

cdef class pyBLSResult:
    cdef readonly double P
    cdef readonly double dchi2
    cdef readonly double mag0
    cdef readonly double dmag
    cdef readonly double t0
    cdef readonly double dur

    def __cinit__(self, pyBLSAnalyzer blsa, size_t index):
        self.P = blsa.P[index]
        self.dchi2 = blsa.dchi2[index]
        self.mag0 = blsa.mag0[index]
        self.dmag = blsa.dmag[index]
        self.t0 = (blsa.t0[index] + blsa.dur[index] / 2) % blsa.P[index]
        self.dur = blsa.dur[index]
    
    def __str__(self):
        return (
            f"pyBLSResult(P={self.P}, dchi2={self.dchi2}, mag0={self.mag0}, dmag={self.dmag}, "
            f"t0={self.t0}, dur={self.dur}, snr={self.snr})"
        )
    
    @property
    def r(self):
        return (self.dmag / self.mag0)**0.5

    def snr(self, pyDataContainer phot):
        cdef bool_t[:] mask = self.get_transit_mask(phot.rjd)
        cdef double err_in = np.sum(phot.err[mask]**-2)**-0.5
        return self.dmag / err_in

    def get_transit_mask(self, double[:] t):
        return (abs((np.array(t) - self.t0 + self.P / 2) % self.P - self.P / 2) < self.dur / 2)

cdef int convert_duration_mode(str duration_mode):
    """
    Converts a string representation of a duration mode to its integer counterpart.
    """
    cdef dict allowed_duration_modes = {'': 0,
                                        'constant': 1,
                                        'fractional': 2,
                                        'physical': 3}
    assert (
        duration_mode in allowed_duration_modes
        ), f"duration_mode must be one of: {allowed_duration_modes.keys()}"
    
    return allowed_duration_modes[duration_mode]


import gerbls
import numpy as np
from .exofunc import divide_into_chunks
from scipy.signal import savgol_filter
from typing import Union


def clean_savgol(phot: gerbls.pyDataContainer,
                 N_flares: int = 3,
                 sigma_clip: float = 5.,
                 window_length: float = 1.,
                 chunk_length: float = 0.,
                 verbose: bool = True):
    """
    Clean the data using a Savitsky-Golay filter.
    Also performs sigma clipping and flare rejection.
    Cubic splines will be used to interpolate over any masked data.

    Parameters
    ----------
    phot : gerbls.pyDataContainer
        Input data to be cleaned.
    N_flares : int, optional
        Number of iterations to detect flares, by default 3
    sigma_clip: float, optional
        Whether to remove outliers more than X sigma from the initial baseline, by default 5. To turn off, enter 0.
    window_length: float, optional
        Window length in time units (days) for the Savitsky-Golay filter, by default 1.
    chunk_length: float, optional
        If given, divide the data into chunks first where the time gaps in data are longer than this length,
        and apply the filter independently to each chunk, by default 0.
    verbose: bool, optional
        Whether to print any text/warnings, by default True

    Returns
    -------
    gerbls.pyDataContainer
        Cleaned data.
    np.array
        Mask corresponding to valid cleaned data.
    """

    # Mask corresponding to valid cleaned data
    cmask = np.ones(phot.size, dtype=bool)

    # Determine window size (=1 day) based on median time cadence
    filter_width = int(window_length / np.median(np.diff(phot.rjd)))
    if filter_width % 2 == 0:
        filter_width += 1

    # Subroutine that fits the filter to masked data
    def fit(mask):

        mag_data = np.copy(phot.mag)
        mag_fit = np.zeros(phot.size)

        # Use cubic splines to interpolate over unmasked data
        mag_data[~mask] = np.interp(phot.rjd[~mask], phot.rjd[mask], phot.mag[mask])

        # Divide data into chunks with gaps over 0.2 days
        for a, b in divide_into_chunks(phot.rjd, chunk_length):
            if b - a < filter_width:
                filter_width_ = b - a - (1 if (b - a) % 2 == 0 else 2)
                if verbose:
                    print(
                        f"Warning: Savgol window length shortened {filter_width} => {filter_width_}")
            else:
                filter_width_ = filter_width
            mag_fit[a:b] = savgol_filter(mag_data[a:b], filter_width_, 3)

        return mag_fit[mask]

    if sigma_clip > 0:
        mag0 = fit(cmask)
        cmask = (abs(phot.mag - mag0) <= phot.err * sigma_clip)

    for _ in range(N_flares):
        phot_ = phot.mask(cmask)
        cmask[cmask] = ~find_flares(phot_, fit(cmask))

    cphot = phot.mask(cmask)
    cphot.remove_planet(fit(cmask))

    return cphot, cmask


def find_flares(phot: gerbls.pyDataContainer,
                mag0: Union[np.ndarray, float] = 1):
    """
    Returns a mask with detected flares.
    A flare is at least 4 consecutive positive 2-sigma deviants.

    Parameters
    ----------
    phot : cinject.pyDataContainer
        Input data.
    mag0 : float or np.ndarray, optional
        Flux baseline to be subtracted before flare detection, by default 1.

    Returns
    -------
    np.ndarray[bool]
        Mask denoting flare data.
    """

    mag = phot.mag - mag0
    mask_ = (mag / phot.err >= 2)   # Mask for single points
    cs = np.cumsum(mask_, dtype=int)

    # Indices for first points of flares
    i_start = np.where(cs[4:] - cs[:-4] >= 4)[0] + 1

    # Mask for flares
    mask = np.zeros(phot.size, dtype=bool)

    j = 0
    for i in i_start:
        if i < j:
            continue
        j = i + 4
        while i > 0 and mag[i-1] > 0:
            i -= 1
        while j < phot.size and mag[j] > 0:
            j += 1
        mask[i:j] = True

    return mask



# distutils: language = c++
# Cython bindings for imearth

from libcpp cimport bool as bool_t
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
#from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector

cdef extern from "cpp/ffafunc.hpp":
    unique_ptr[DataContainer] resample_uniform(DataContainer, double)

cdef extern from "cpp/model.hpp":
    cdef cppclass BLSModel:
        int max_duration_mode
        double max_duration_factor
        vector[double] chi2_dmag
        vector[double] chi2_dt
        vector[double] chi2_mag0
        vector[double] chi2_t0
        vector[double] dchi2
        vector[double] freq
        vector[size_t] N_bins
        
        size_t N_freq()
        void run(bool_t)
    
    cdef cppclass BLSModel_bf(BLSModel):
        BLSModel_bf(
            DataContainer, double, double, Target*, double, double, size_t, int, double, double)
        BLSModel_bf(DataContainer, vector[double], Target*, double, size_t, int, double, double)
    
    cdef cppclass BLSModel_FFA(BLSModel):
        double t_samp
        unique_ptr[DataContainer] rdata
        
        vector[double] periods
        vector[size_t] widths
        vector[size_t] foldbins
        vector[double] snr
        vector[size_t] t0
        BLSModel_FFA(DataContainer, double, double, Target*, int, double, double)
        
        void run_double(bool_t)

cdef extern from "cpp/physfunc.hpp":
    double gdraw(double, double, double)
    double get_aR_ratio(double, double, double)
    double get_inc(double, double, double, double)
    void get_phase_range(double, double*, double*)
    double get_transit_dur(double, double, double, double)
    double grand(double, double)

cdef extern from "cpp/structure.hpp":
    cdef cppclass DataContainer:
        int* sec
        double* rjd
        double* mag
        double* err
        size_t size
        
        unique_ptr[DataContainer] clean(double, bool_t*, int)
        unique_ptr[DataContainer] clean_hw(double, bool_t*, int)
        unique_ptr[bool_t] find_flares(const double*)
        unique_ptr[bool_t] find_flares()
        void imprint(double*, size_t)
        unique_ptr[DataContainer] phase_folded(double, double)
        vector[double] running_median(double)
        vector[double] running_median_eval(double, double*, size_t)
        vector[double] running_median_per(double, double)
        void set(double*, double*, double*, size_t)
        #vector[double] splfit(double, int)
        #vector[double] splfit_eval(int, double*, size_t)
        #unordered_map[int, unique_ptr[DataContainer]] split_by_sector()
        void store(double*, double*, double*, size_t)

    cdef cppclass Target:
        double M
        double R
        double L
        double u1
        double u2
        double L_comp
        double P_rot
        double P_rot2
        
        double logg()
        double Teff()


import gerbls
import numpy as np

def test_bls_basic(phot_test):
    """
    Test for the BLS basic usage function.
    """

    correct_P = 1.37

    results = gerbls.run_bls(phot_test.rjd, phot_test.mag, phot_test.err, 
                             min_period=0.4, max_period=10., t_samp=10/60/24)
    
    # Check the validity of BLS results
    assert 'P' in results
    assert len(results['P']) > 0
    assert not np.any(np.isnan(results['P']))
    assert not np.any(np.isnan(results['dchi2']))
    assert not np.any(np.isnan(results['t0']))
    assert not np.any(np.isnan(results['dur']))
    assert not np.any(np.isnan(results['mag0']))
    assert not np.any(np.isnan(results['dmag']))
    
    # Check whether the correct period was recovered
    assert abs(results['P'][np.argmax(results['dchi2'])] - correct_P) < 0.001

def test_bls_bf(phot_test):
    """
    Test for the brute-force BLS.
    """

    correct_P = 1.37

    bls = gerbls.pyBruteForceBLS()
    bls.setup(phot_test, 0.4, 10., t_bins=10/60/24)
    bls.run()

    blsa = gerbls.pyBLSAnalyzer(bls)
    assert len(blsa.P) > 0

    best_model = blsa.generate_models(1)[0]
    assert abs(best_model.P - correct_P) < 0.001

def test_bls_fast(phot_test):
    """
    Test for the fast-folded BLS.
    """

    correct_P = 1.37

    bls = gerbls.pyFastBLS()
    bls.setup(phot_test, 0.4, 10., t_samp=10/60/24)
    bls.run()

    blsa = gerbls.pyBLSAnalyzer(bls)
    assert len(blsa.P) > 0

    best_model = blsa.generate_models(1)[0]
    assert abs(best_model.P - correct_P) < 0.001


from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy as np

setup(name = "_gerbls",
      ext_modules = cythonize([Extension("_gerbls",
                                         ["src/gerbls/gerbls.pyx",
                                          "src/gerbls/cpp/ffafunc.cpp",
                                          "src/gerbls/cpp/model.cpp",
                                          "src/gerbls/cpp/physfunc.cpp",
                                          "src/gerbls/cpp/structure.cpp"],
                                         include_dirs=[np.get_include()],
                                         define_macros=[("NPY_NO_DEPRECATED_API", 
                                                         "NPY_1_7_API_VERSION")],
                                         extra_compile_args = ["-O3",
                                                               "-std=c++0x",
                                                               "-march=native",
                                                               "-fassociative-math",
                                                               "-fno-math-errno",
                                                               "-ffinite-math-only",
                                                               "-fno-rounding-math",
                                                               "-fno-signed-zeros",
                                                               "-fno-trapping-math"])], 
                              annotate=False,
                              compiler_directives={"embedsignature": True}),
      zip_safe = False,
      )
